/**
 * $Id: Actor.java,v 1.70 2003/10/06 20:56:42 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;

import java.io.*;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;
import java.util.Hashtable;

import salsa.language.exceptions.CurrentContinuationException;
import salsa.messaging.*;
import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;
import salsa.naming.MalformedUANException;
import salsa.naming.MalformedUALException;

import salsa.resources.StandardOutput;
import salsa.resources.StandardError;

public class Actor extends Thread implements java.io.Serializable {
	public UAN __uan;
	public UAL __ual;

	public Vector mailbox = new Vector();

	protected boolean migrated = false;
	protected boolean stayAlive = true;
	protected TokenInput currentContinuation = null;
	private Message currentMessage = null;

	public transient TransportService transportService;
	public transient NamingService namingService;
	public transient TheaterService theaterService;

	public static StandardOutput standardOutput = StandardOutput.getInstance();
	public static StandardError standardError = StandardError.getInstance();

	//this is a vector of all the messages generated by a message
	//processed by this actor.  These are all sent when the message
	//has completed being processed
	public Vector messages = new Vector();

	public Actor() {
		//get the implementations for the required services for an Actor
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		__uan = null;
		__ual = theaterService.generateUAL();

		setName(getClass().getName());
		start();
	}

	/** Die kills the actor when its mailbox has been
	  * emptied
	  */
	public void die()	{ stayAlive = false; }
	/** Destroy immediately kills the actor without processing any
	  * more messages
	  */
	public void destroy()	{ migrated = true; }

	public UAN getUAN()	{ return __uan; }
	public UAL getUAL()	{ return __ual; }


	private Hashtable unresolvedTokens = new Hashtable();
	private Vector pendingTokens = new Vector();

	public synchronized void resolveToken(String id, Object value) {
		TokenOutput target = (TokenOutput)unresolvedTokens.get(id);

		if (target != null) {
			target.replace(value);

			if (target.requiredValues < 1) unresolvedTokens.remove(id);
		} else {
			//The token has reached the actor before the message it was intended for,
			//add this to a vector of tokens waiting for messages.
			pendingTokens.add( new PendingToken(id, value) );
		}
	}

	public void addUnresolvedToken(TokenOutput unresolvedToken) {
		unresolvedTokens.put( unresolvedToken.getId(), unresolvedToken );
	}

	public void MessageTargetNotFound(Message message) {
		ActorReference target = message.getTarget();
		UAN uan = target.getUAN();

		if (uan == null) {
			System.err.println("Message sending error:");
			System.err.println("\tMessage: " + message.getDescriptor());
			System.err.println("\tFrom: " + toString());
			System.err.println("\tTo: " + target.toString());
			System.err.println("\tException: Could not find the target actor.");
			System.err.println("\t\t   It has no UAN and was not found at the theater of its UAL.");
			return;
		} else {
			UAL ual = namingService.resolve(uan);
			if (ual == null) {
				System.err.println("Message sending error:");
				System.err.println("\tMessage: " + message.getDescriptor());
				System.err.println("\tFrom: " + toString());
				System.err.println("\tTo: " + target.toString());
				System.err.println("\tException: Could not find the target actor.");
				System.err.println("\t\t   The naming service could not resolve its UAL from its UAN.");
				return;
			}
			theaterService.setEntry( uan, ual );

			if ( (target.getActor() == null) && target.getUAL().getLocation().equals(theaterService.getLocation()) ) {
				System.err.println("Message sending error:");
				System.err.println("\tMessage: " + message.getDescriptor());
				System.err.println("\tFrom: " + toString());
				System.err.println("\tTo: " + target.toString());
				System.err.println("\tException: Could not find the target actor.");
				System.err.println("\t\t   The naming service says that the actor should be at the local theater, but it is not present.");
				return;
			}

			target.send(message);
		}
	}


	public synchronized void send(Message message) {
		RunTime.receivedMessage();

		if (((__uan != null) && (theaterService.getEntry(__uan) == this)) ||
		    ((__ual != null) && (theaterService.getEntry(__ual) == this))) {
			if (message.getRequiredTokens() > 0) {
				Object[] arguments = message.getArguments();
				for (int i = 0; i < arguments.length; i++) {
					if (arguments[i] instanceof TokenOutput) {
						addUnresolvedToken( (TokenOutput)arguments[i] );
					}
				}
				if (message.getInput() instanceof TokenOutput) {
					addUnresolvedToken( (TokenOutput)message.getInput() );
				}

				//try and resolve the pending tokens
				for (int i = 0; i < pendingTokens.size(); i++) {
					PendingToken pendingToken = (PendingToken)pendingTokens.get(i);
					TokenOutput target = (TokenOutput)unresolvedTokens.get( pendingToken.id );

					if (target != null) {
						target.replace( pendingToken.value );
						if (target.requiredValues < 1) unresolvedTokens.remove( pendingToken.id);
						pendingTokens.remove(i);
						i--;
					}
				}

			} else {
				if (message.getProperty() != null && message.getProperty().equals("priority")) {
					mailbox.insertElementAt(message, 0);
				} else {
					mailbox.addElement(message);
				}
				notify();
			}
		} else {
			message.getTarget().send(message);
			notify();
		}
	}

	public synchronized Message getMessage() {
		if (mailbox.isEmpty()) {
			try {
				wait();      // The lock of Mailbox means it is empty.
			} catch (InterruptedException ie){
				System.err.println("Error from within salsa.language.Actor: " + toString());
				System.err.println("\tError getting a new message:");
				System.err.println("\t" + ie);
			}
		}
		Message nextMessage = (Message)mailbox.firstElement();
		mailbox.removeElementAt(0);

		return nextMessage;
	}

	public void run() {
		while ((!migrated) && (stayAlive || !mailbox.isEmpty()) ) {
			currentMessage = getMessage();

			currentContinuation = currentMessage.getOutput();
			process(currentMessage);
			RunTime.finishedProcessingMessage();
			currentContinuation = null;
		}
		if (currentMessage.getMethodName().equals("destroy")) {
			//remove this actor from the registry
			theaterService.removeEntry(__uan);
			theaterService.removeEntry(__ual);
			//namingService.delete(__uan);
		}
	}

	public String toString() {
		String name = "";
		name +=  getClass().getName() + ": ";
		if (__uan != null) name += __uan.toString() + ", ";
		else name += "null, ";
		if (__ual != null) name += __ual.toString();
		else name += "null";
		return name;
	}


	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.

	 */
	public void bind(String uan, String ual) throws MalformedUANException, MalformedUALException {
		bind(new UAN(uan), new UAL(ual));
	}

	public synchronized void bind(UAN uan, UAL ual) {
		//bind this actor, then migrate it to the appropriate target.
		bind(uan);
		migrate(ual);
	}

	/**
	 * This method binds the actor to the specified name and adds the actor to
	 * an unspecified location at the current theater. An on-the-fly theater is created
	 * if one does not already exist.
	 * 
	 * @author stepha
	 * @param uan or ual
	 */
	public void bind( String name ) throws MalformedUANException {
		bind( new UAN( name ) );
	}
	
	public synchronized void bind( UAN uan ) {
		this.__uan = uan;

		theaterService.setEntry(__ual, __uan);
		theaterService.setEntry(__uan, this);

		// Update the naming service.
		namingService.add( __uan, __ual );
		RunTime.receivedUniversalActor();
	}

	public synchronized void reName ( String identifier ) {
		UAL ual = new UAL( "rmsp://" + __ual.getHost() + ":" + __ual.getPort() + identifier );

		if (__uan != null) {
			theaterService.setEntry(ual, __uan);
		} else {
			theaterService.setEntry(__ual, ual);
			theaterService.setEntry(ual, this);
		}

		this.__ual = ual;

		// Update the naming service.
		if (__uan != null) namingService.add( __uan, __ual );

		RunTime.receivedUniversalActor();
	}


        public boolean equals(Object o) {
                if (o instanceof ActorReference) {
                        ActorReference ref = (ActorReference)o;
                        //make sure all the references on these actors
                        //are correct
			ref.getActor();
                        
                        if (__uan != null && ref.uan != null) return __uan.equals(ref.uan);
                        else if (__ual != null && ref.ual != null) return __ual.equals(ref.ual);
                } if (o instanceof Actor) {
                        Actor ref = (Actor)o;

                        if (__uan != null && ref.getUAN() != null) return __uan.equals(ref.getUAN());
                        else if (__ual != null && ref.getUAL() != null) return __ual.equals(ref.getUAN());
                }
                return false;
        }

	/**
	 * This method moves the actor from one theater to another.
	 */
	public void migrate(String newUAL) throws MalformedUALException {
		migrate(new UAL(newUAL));
	}

	public synchronized void migrate(UAL ual) {
		// Update the name server.
		if (__uan == null) {
			System.err.println("Actor migration error:");
			System.err.print("\tImproper migration request from: ");
			if (currentMessage.getSender() == null) System.err.println("unknown");
			else System.err.println(currentMessage.getSender().toString());
			System.err.println("\tBy message: " + currentMessage.getDescriptor());
			System.err.println("\tException: Cannot migrate an actor which does not have a UAN.");
			return;
		}

		namingService.update(__uan, ual);

		this.__ual = ual;
		// Perform the migration.
		transportService.migrate(this, __ual);

		if ( theaterService.getEntry( __uan ).equals(this) ) {
			//make sure we haven't migrated to a theater we were already at
			theaterService.setEntry( __uan, __ual );
		}			
		migrated = true;
	}



	/** The following are used to serialize and deserialize an Actor
	 */
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		out.flush();
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		
		// Get service references from the ServiceFactory.
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		theaterService.setEntry(__uan, this);
		theaterService.setEntry(__ual, __uan);

		standardOutput = StandardOutput.getInstance();
		standardError = StandardError.getInstance();

		setName(getClass().getName());
	}



	/**
		The process method is responsible for determining which
		child class method should be invoked for a specific message.
		The process method first invokes the chosen method, updates any
		tokens and then sends any continuation the message contains.

		Note that Class.getMethod() is not used for method resolution.
		The getMethod() method will not match methods for which a
		parameter type is unknown. This algorithm will in the case that
		only one method name matches and has the same number of arguments.

		Instead the following algorithm is applied recursively to all
		this class and all superclasses until the Actor class is
		reached. Null arguments are given Object parameter types.

		1.) Check for an exact match between method name and parameter
		    type class names.

		2.) Check for an exact match between method name, parameter
		    number and for assignability between parameter types.

		3.) If above two checks fail, repeat for superclass if Actor
		    is the superclass then fail.
	*/
	public void process(Message message) {
		Method match = getInvocationTarget(message);
		Object token = null;

		if (match != null) {
			try {
				token = match.invoke(this, message.getArguments());
			} catch (InvocationTargetException e) {
				Throwable wrapped = e.getTargetException();

				if ( !(wrapped instanceof CurrentContinuationException) ) {
					System.err.println("Message processing exception:");
					if (message.getSender() != null) {
						System.err.println("\tSent by: " + message.getSender().toString());
					} else System.err.println("\tSent by: unknown");
					System.err.println("\tBy actor: " + toString());
					System.err.println("\tInvocation of " + message.getMethodName() + " generated a " +
							    wrapped.getClass().getName() + " exception: " + wrapped.getMessage() );
				}
			} catch (IllegalAccessException e) {
				System.err.println("Message processing exception: ");
				System.err.println("\tBy actor: " + toString());
				if (message.getSender() != null) {
					System.err.println("\tSent by: " + message.getSender().toString());
				} else System.err.println("\tSent by: unknown");
				System.err.println("\tIllegal access exception generated by the invocation of " + message.getMethodName() + ": " + e.getMessage() );

			}
		}

                transportService.send( messages );
                messages = new Vector();
                
                // Send the continuation.
                if (currentContinuation != null) {
                        //message.getTarget is a reference to this actor, and is needed
                        //to send the return value as a "resolveToken" message to its  
                        //target actor(s).
                        currentContinuation.resolve(message.getTarget(), token);
                }
                RunTime.finishedProcessingMessage();
	}

	public Method getInvocationTarget(Message message) {
		String methodName = message.getMethodName();
		Object[] messageArguments = message.getArguments();

		Class[] mpt = new Class[ messageArguments.length ];

		for (int i = 0; i < mpt.length; ++i) {
			// Check to see if the argument is null. If so set its type
			// to be object.
			if (messageArguments[i] == null) {
				mpt[i] = null;
			} else {
				// Otherwise get the argument's type.
				mpt[i] = messageArguments[i].getClass();
			}
		}

		// Check the current Class and then all super classes for a matching method.
		Class state = getClass();
		Method match = null;

		while (!state.getName().equals( "java.lang.Object" )) {
			// Examine each method declared in this class.
			Method[] declared = state.getDeclaredMethods();

			for (int i = 0; i < declared.length; ++i) {
				Method candidate = declared[i];

				//cpt stands for candidateParameterTypes, for the same reason mpt stands for
				//messageParameterTypes
				Class[] cpt = candidate.getParameterTypes();

				// Check for an exact match between method name and parameter number.
				if (candidate.getName().equals( methodName ) && cpt.length == mpt.length) {
					// Check if all the parameter types match exactly.
					int j;
					for (j = 0; j < mpt.length; ++j) {
						if ( (mpt[j] == null) || !mpt[j].getName().equals( cpt[j].getName() )) {
							break;
                  				}
               				}

               				// Matched all successfully.
               				if (j == mpt.length) {
                  				match = candidate;
                 				break;
               				}
           			}
			}

         		// See if a match has been found at this class.
         		if (match != null) {
            			break;
         		}

         		for (int i = 0; i < declared.length; ++i) {
            			Method candidate = declared[i];

				//Again, cpt stands for CandidateParameterTypes, to save
				//valuable hard drive space
            			Class[] cpt = candidate.getParameterTypes();
            			// Check for an exact match between method name and parameter number.

            			if (candidate.getName().equals( methodName ) && cpt.length == mpt.length) {

               				// Check if the candidate parameter types are assignable from
               				// the message parameter types.
               				int j;
               				for (j = 0; j < mpt.length; ++j) {
                  				if ( !((mpt[j] == null) && !cpt[j].isPrimitive()) && !cpt[j].isAssignableFrom( mpt[j] ) ) {
							// If the candidate parameter type is a primitive type check to see if the
							// message parameter is any of the nine primitive wrapper classes
							// needs to be changed,  check and see if cpt is primitive and mpt is wrapper
							// If the candidate parameter type is a primitive type check to see if the
							// message parameter is any of the nine primitive wrapper classes

							if (!((cpt[j].isPrimitive() || cpt[j].isArray()) &&
							   (	cpt[j] == Integer.TYPE && mpt[j] == Integer.class ||
								cpt[j].getName().equals( "[I" ) && mpt[j].getName().equals( "[Ljava.lang.Integer;" ) ||

								cpt[j] == Boolean.TYPE && mpt[j] == Boolean.class ||
								cpt[j].getName().equals( "[Z" ) && mpt[j].getName().equals( "[Ljava.lang.Boolean;" ) ||

								cpt[j] == Character.TYPE && mpt[j] == Character.class ||
								cpt[j].getName().equals( "[C" ) && mpt[j].getName().equals( "[Ljava.lang.Character;" ) ||

								cpt[j] == Byte.TYPE && mpt[j] == Byte.class ||
								cpt[j].getName().equals( "[B" ) && mpt[j].getName().equals( "[Ljava.lang.Byte;" ) ||

								cpt[j] == Short.TYPE && mpt[j] == Short.class ||
								cpt[j].getName().equals( "[S" ) && mpt[j].getName().equals( "[Ljava.lang.Short;" ) ||

								cpt[j] == Long.TYPE && mpt[j] == Long.class ||
								cpt[j].getName().equals( "[J" ) && mpt[j].getName().equals( "[Ljava.lang.Long;" ) ||

								cpt[j] == Float.TYPE &&	mpt[j] == Float.class ||
								cpt[j].getName().equals( "[F" ) && mpt[j].getName().equals( "[Ljava.lang.Float;" ) ||

								cpt[j] == Double.TYPE && mpt[j] == Double.class ||
								cpt[j].getName().equals( "[D" ) && mpt[j].getName().equals( "[Ljava.lang.Double;" ) ||

								cpt[j] == Void.TYPE && mpt[j] == Void.class
							   ))) {
								break;
							}

							// If the matching candidate parameter is a primitive array type and
							// the message parameter is a wrapper type convert the message parameter
							// to a primitive array.
							if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Integer;" )) {

								// Copy the values.
								int[] array = new int[ ((Integer[])messageArguments[j]).length ];
								for (int w=0; w < ((Integer[])messageArguments[j]).length; w++) {
									array[w] = ((Integer[])messageArguments[j])[w].intValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Boolean;" )) {

								// Copy the values.
								boolean[] array = new boolean[ ((Boolean[])messageArguments[j]).length ];
								for (int w=0; w < ((Boolean[])messageArguments[j]).length; ++w) {
									array[w] = ((Boolean[])messageArguments[j])[w].booleanValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Character;" )) {

								// Copy the values.
								char[] array = new char[ ((Character[])messageArguments[j]).length ];
								for (int w=0; w < ((Character[])messageArguments[j]).length; ++w) {
									array[w] = ((Character[])messageArguments[j])[w].charValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Byte;" )) {

								// Copy the values.
								byte[] array = new byte[ ((Byte[])messageArguments[j]).length ];
								for (int w=0; w < ((Byte[])messageArguments[j]).length; ++w) {
									array[w] = ((Byte[])messageArguments[j])[w].byteValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Short;" )) {

								// Copy the values.
								short[] array = new short[ ((Short[])messageArguments[j]).length ];
								for (int w=0; w < ((Short[])messageArguments[j]).length; ++w) {
									array[w] = ((Short[])messageArguments[j])[w].shortValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Long;" )) {

								// Copy the values.
								long[] array = new long[ ((Long[])messageArguments[j]).length ];
								for (int w=0; w < ((Long[])messageArguments[j]).length; ++w) {
									array[w] = ((Long[])messageArguments[j])[w].longValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Float;" )) {

								// Copy the values.
								float[] array = new float[ ((Float[])messageArguments[j]).length ];
								for (int w=0; w < ((Float[])messageArguments[j]).length; ++w) {
									array[w] = ((Float[])messageArguments[j])[w].floatValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Double;" )) {

								// Copy the values.
								double[] array = new double[ ((Double[])messageArguments[j]).length ];
								for (int w=0; w < ((Double[])messageArguments[j]).length; ++w) {
									array[w] = ((Double[])messageArguments[j])[w].doubleValue();
								}

								messageArguments[j] = array;
							}
						}
					}

					// Matched all successfully.
					if (j == mpt.length) {
						match = candidate;
						break;
					}
				}
			}

			// See if a match has been found at this class.
			if (match != null) {
				break;
			}

			// Examine the parent class next.
			state = state.getSuperclass();
		}

		if (match != null) {
			//Before invoking the method, check to see if the remote actor has access
			//to the method.
			int modifiers = ((Member)match).getModifiers();

			if ( Modifier.isPrivate(modifiers) ) {
				ActorReference sender = message.getSender();
				//if sender is null, it was sent from the SALSA system
				//and we can invoke it.
				if ( sender != null && !this.getUAL().equals(sender.getUAL()) ) {
					System.err.println("Message invocation exception: ");
					System.err.println("\tMessage: " + methodName);
					if (message.getSender() != null) {
						System.err.println("\tSent by: " + message.getSender().toString());
					} else System.err.println("\tSent by: unknown");
					System.err.println("\tBy actor: " + toString());
					System.err.println("\tIllegal Access: " + methodName + " is private, and cannot be invoked by the source of this message.");
					return null;
				}
			} else if ( Modifier.isProtected(modifiers) ) {
				ActorReference sender = message.getSender();
				//if sender is null, it was sent from the SALSA system
				//and we can invoke it.

				if ( sender != null ) {
					Package senderPackage = sender.getClass().getPackage();
					Package currentPackage = this.getClass().getPackage();

					if ( senderPackage != null && senderPackage.equals(currentPackage) ) {
						System.err.println("Message invocation exception: ");
						System.err.println("\tMessage: " + methodName);
						if (message.getSender() != null) {
							System.err.println("\tSent by: " + message.getSender().toString());
						} else System.err.println("\tSent by: unknown");
						System.err.println("\tBy actor: " + toString());
						System.err.println("\tIllegal Access: " + methodName + " is protected, and cannot be invoked by the source of this message.");
						return null;
					}
				}
			}
		}

		if (match == null) {
			System.err.println("Message processing exception: ");
			System.err.println("\tBy actor: " + toString());
			if (message.getSender() != null) {
				System.err.println("\tSent by: " + message.getSender().toString());
			} else System.err.println("\tSent by: unknown");
			System.err.println("\tUnknown method: " + message.getDescriptor());
		}

		message.setArguments(messageArguments);
		return match;
	}
}
