package wwc.messaging;

import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.UnknownHostException;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Vector;

import salsa.language.*;
import salsa.language.exceptions.*;
import salsa.naming.*;
import salsa.messaging.*;

/**
 * This class is the reference implementation of the MessagingService. The
 * service is implemented as a RMSP client.
 * 
 * This implementation uses the NamingService interface and the salsa.language.
 * ServiceFactory for UAN resolution.
 * 
 * Author: Travis Desell
 */
public class PersistentMessenger implements TransportService {
	public PersistentMessenger()		{ super(); }
	private Hashtable sockets		= 	new Hashtable();
	private Hashtable incomingStreams	= new Hashtable();
	private Hashtable outgoingStreams	= new Hashtable();

	private synchronized Socket getSocket(UAL target){
		String key = target.getHost() + target.getPort();

		Socket socket = (Socket)sockets.get(key);

		if (socket == null || !socket.isConnected()) {
                                System.out.println("GetOutputStream: Generating a new socket");
                                //we need to generate a new socket, and put it in the hashtable
                                try {
                                        socket = new Socket(target.getHost(), target.getPort());
										socket.setKeepAlive(true);
                                } catch (UnknownHostException e) {
                                        System.err.println("Unknown host: " + target.toString());
                                        System.err.println("\terror: " + e);
                                } catch (IOException e) {
                                        System.err.println("Error in creating a socket for an output stream: " + e);
                                        System.err.println("\twith target: " + target.toString());
                                }
                                sockets.put(key, socket);
		}

		return socket;            

	}

	private synchronized ObjectOutputStream getOutputStream(UAL target) {
		String key = target.getHost() + target.getPort();
		ObjectOutputStream outputStream = (ObjectOutputStream)outgoingStreams.get(key);

		if (outputStream == null) {
			//we haven't sent anything to this foreign theater yet, so set up an output stream for it
			Socket socket = (Socket)sockets.get(key);

			if (socket == null || !socket.isConnected()) {
				System.out.println("GetOutputStream: Generating a new socket");
				//we need to generate a new socket, and put it in the hashtable
				try {
					socket = new Socket(target.getHost(), target.getPort());
					socket.setKeepAlive(true);
				} catch (UnknownHostException e) {
					System.err.println("Unknown host: " + target.toString());
					System.err.println("\terror: " + e);
				} catch (IOException e) {
					System.err.println("Error in creating a socket for an output stream: " + e);
					System.err.println("\twith target: " + target.toString());
				}
				sockets.put(key, socket);
			}
			try {
				outputStream = new ObjectOutputStream(socket.getOutputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			outgoingStreams.put(key, outputStream);
		}
		return outputStream;
	}

	private synchronized ObjectInputStream getInputStream(UAL target) {
		String key = target.getHost() + target.getPort();
		ObjectInputStream inputStream = (ObjectInputStream)incomingStreams.get(key);

		if (inputStream == null) {
			//we haven't sent anything to this foreign theater yet, so set up an output stream for it
			Socket socket = (Socket)sockets.get(key);

			if (socket == null || !socket.isConnected()) {
				System.out.println("GetInputStream: Generating a new socket");
				//we need to generate a new socket, and put it in the hashtable
				try {
					socket = new Socket(target.getHost(), target.getPort());
					socket.setKeepAlive(true);
				} catch (UnknownHostException e) {
					System.err.println("Unknown host: " + target.toString());
					System.err.println("\terror: " + e);
				} catch (IOException e) {
					System.err.println("Error in creating a socket for an output stream: " + e);
					System.err.println("\twith target: " + target.toString());
				}
				sockets.put(key, socket);
			}
			try {
				inputStream = new ObjectInputStream(socket.getInputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + target.toString());
			}
			incomingStreams.put(key, inputStream);
		}
		return inputStream;
	}

	public void migrate(Actor state, UAL target) {
		//ObjectOutputStream outStream;
		//Socket socket = getSocket( target );
		try {
                        //outStream = new ObjectOutputStream(socket.getOutputStream());	
			ObjectOutputStream outStream = getOutputStream( target );
			outStream.writeObject( state );
			outStream.flush();
		} catch (IOException e) {
			System.err.println("Error migrating actor: (" + state.getUAN().toString() + "," + state.getUAL().toString() + ")");
			System.err.println("\tto target: " + target.toString());
			System.err.println("\terror: " + e);
		}

		// Wait for the Acknowledgement that the actor is situated on the remote theater.
		try {
			RMSPAcknowledgement ack = (RMSPAcknowledgement)getInputStream( target ).readObject();
			//ObjectInputStream inStream = new ObjectInputStream(socket.getInputStream());
			//RMSPAcknowledgement ack = (RMSPAcknowledgement)inStream.readObject();
		} catch (IOException e) {
			System.err.println("Error sending migration acknowledgement to: " + target.toString());
			System.err.println("\terror: " + e);
		} catch (ClassNotFoundException e) {
			System.err.println("RMSPAcknowledgemet class not found, please update SALSA package.");
			System.err.println("\terror: " + e);
		}
	}

	public void send(Message message, UAL target) {
		// Determine the destination of this message.
		System.out.println("Send: Using persistent messenger");
		if (target == null) {
			UAN uan = message.getTarget().getUAN();

			if (uan != null) {
				Object entry = ServiceFactory.getTheater().getEntry(uan);
				if ((entry instanceof UAL) || (entry == null)) {
					// Resolve the location.
					target = ServiceFactory.getNaming().resolve( uan );
					ServiceFactory.getTheater().setEntry(uan, target);
				} else {
					((Actor)entry).send(message);
					return;
				}
			} else {
				throw new SalsaException( "UAN and UAL of message are null, cannot send message: " + message.getName() );
			}
		}

		try {
			ObjectOutputStream outStream = getOutputStream( target );
			//Socket socket = getSocket( target );
			 
			//ObjectOutputStream outStream = new ObjectOutputStream(socket.getOutputStream());  
			outStream.writeObject( message );
		} catch (IOException e) {
			System.err.println("Error sending message: " + message.getName());
			System.err.println("\tto target: " + target.toString());
			System.err.println("\terror: " + e);
		}
	}

	public void send(Vector messages) {
		//used by the compiler to send a vector of messages generated by a message

		for (int i = 0; i < messages.size(); i++) {
			Message current = (Message)messages.get(i);
			current.getTarget().send( current );
		}
	}
}
